from posym.operations import Operation
import numpy as np


class Identity(Operation):
    def __init__(self, label):
        super().__init__(label)

    def __hash__(self):
        return hash((self._label))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def get_measure_modes(self, coordinates, modes, symbols, orientation=None):
        self._measure_mode = [1.0] * len(modes)

        return np.array(self._measure_mode)

    def get_measure_atom(self, coordinates, symbols, orientation=None):
        return len(coordinates)

    def get_measure_xyz(self, orientation=None):
        return 3

    def get_displacements_projection(self, coordinates, symbols, orientation=None):
        return np.identity(3*len(symbols)).reshape(3*len(symbols), len(symbols), 3)#.tolist()

    def get_operated_coordinates(self, coordinates, symbols, orientation=None):
        return np.array(coordinates)

    def get_overlap_func(self, op_function1, op_function2, orientation=None):
        return (op_function1*op_function2).integrate

    def get_measure_pos(self, coordinates, symbols, orientation=None, normalized=True):
        if normalized:
            return 1.0
        else:
            return np.einsum('ij, ij -> ', coordinates, coordinates)

    def apply_rotation(self, orientation):
        pass

    @property
    def operation_matrix_list(self):
        return [np.identity(3)]

    def __mul__(self, other):
        if not other.__class__.__bases__[0] is Operation:
            raise Exception('Product only defined between Operation subclasses')
        else:
            return [Identity(self._label)]




import numpy as np
from functools import lru_cache


@lru_cache(maxsize=100)
def get_submatrix_indices(symbols):
    # separate distance_table in submatrices corresponding to a single symbol
    submatrices_indices = []
    for s in np.unique(symbols):
        submatrices_indices.append([j for j, s2 in enumerate(symbols) if s2 == s])

    return submatrices_indices


def get_permutation_hungarian(distance_table, symbols):
    """
    This function takes distance_table and returns the permutation vector
    that minimizes its trace, using the Hungarian method.
    """
    from scipy.optimize import linear_sum_assignment

    def get_perm_submatrix(sub_matrix):
        row_ind, col_ind = linear_sum_assignment(sub_matrix)
        perm = np.zeros_like(row_ind)
        perm[row_ind] = col_ind
        return perm

    submatrices_indices = get_submatrix_indices(symbols)

    # determine the permutation for each submatrix
    perm_submatrices = []
    for index in submatrices_indices:
        submatrix = np.array(distance_table)[index, :][:, index]
        perm_sub = get_perm_submatrix(submatrix)
        perm_submatrices.append(perm_sub)

    # restore global permutation by joining permutations of submatrices
    global_permutation = np.zeros(len(distance_table), dtype=int)
    for index, perm in zip(submatrices_indices, perm_submatrices):
        index = np.array(index)
        global_permutation[index] = index[perm]

    return list(global_permutation)


def cache_permutation(func):
    cache_dict = {}

    def wrapper_cache(self, operation, coordinates, symbols):

        hash_key = (np.array2string(operation), np.array2string(coordinates), tuple(symbols))
        if hash_key in cache_dict:
            return cache_dict[hash_key]

        cache_dict[hash_key] = func(self, operation, coordinates, symbols)
        return cache_dict[hash_key]

    return wrapper_cache


class Operation:
    def __init__(self, label):
        self._label = label

    @cache_permutation
    def _get_permutation(self, operation, coordinates, symbols):
        operated_coor = np.dot(operation, coordinates.T).T
        symbols = tuple(int.from_bytes(num.encode(), 'big') for num in symbols)

        dot_table = -np.dot(coordinates, operated_coor.T)
        permutation = get_permutation_hungarian(dot_table, symbols)

        return permutation

    def _get_operated_coordinates(self, operation, coordinates, symbols):
        """
        get coordinates operated and permuted

        :param operation: operator matrix representation (3x3)
        :param coordinates: coordinates to be operated
        :param symbols: atomic symbols
        :return: operated and permuted coordinates
        """
        operated_coor = np.dot(operation, coordinates.T).T
        permutation = self._get_permutation(operation, coordinates, symbols)
        permu_coor = operated_coor[permutation]
        return permu_coor

    @property
    def label(self):
        return self._label



from posym.operations import Operation
from scipy.spatial.transform import Rotation as R
from posym.tools import standardize_vector
import numpy as np
import hashlib


def rotation(angle, rotation_axis):

    rotation_vector = angle * np.array(rotation_axis) / np.linalg.norm(rotation_axis)
    rotation = R.from_rotvec(rotation_vector)

    return rotation.as_matrix()


def prepare_vector(positions, vector):
    return np.array(vector) + np.array(positions)


class Rotation(Operation):
    def __init__(self, label, axis, order=1, exp=1):
        super().__init__(label)

        self._axis = standardize_vector(axis)
        self._order = order
        self._exp = exp

    def __hash__(self):

        axis = np.round(self._axis, decimals=6)
        axis[axis == 0.] = 0.

        return hash((self._label,
                     np.array2string(axis),
                     int(self._order),
                     int(self._exp)
                     ))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def get_measure_modes(self, coordinates, modes, symbols, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        measure_mode = []
        # for angle in np.linspace(2*np.pi/self._order, 2*np.pi, self._order)[:-1]:
        for angle in [2 * np.pi / self._order * self._exp, -2 * np.pi / self._order * self._exp]:

            operation = rotation(angle, rotated_axis)
            permu = self._get_permutation(operation, coordinates, symbols)

            measure_mode_list = []
            for mode in modes:
                operated_mode = np.dot(operation, np.array(mode).T).T
                permu_mode = np.array(operated_mode)[permu]
                measure_mode_list.append(np.trace(np.dot(mode, permu_mode.T))/np.linalg.norm(mode))

            measure_mode.append(measure_mode_list)

        measure_mode_total = np.average(measure_mode, axis=0)

        return measure_mode_total

    def get_measure_atom(self, coordinates, symbols, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        angle = 2 * np.pi / self._order * self._exp
        operation = rotation(angle, rotated_axis)

        permu = self._get_permutation(operation, coordinates, symbols)
        measure_atoms = np.array([1 if i == p else 0 for i, p in enumerate(permu)])

        return np.sum(measure_atoms)

    def get_measure_xyz(self, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        measure_mode = []
        # for angle in np.linspace(2*np.pi/self._order, 2*np.pi, self._order)[:-1]:
        for angle in [2 * np.pi / self._order * self._exp, -2 * np.pi / self._order * self._exp]:

            operation = rotation(angle, rotated_axis)

            measure_mode_list = []
            for axis in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]:
                operated_axis = np.dot(operation, axis)
                measure_mode_list.append(np.dot(axis, operated_axis))

            measure_mode.append(measure_mode_list)

        measure_mode_total = np.average(measure_mode, axis=0)

        return np.sum(measure_mode_total)

    def get_displacements_projection(self, coordinates, symbols, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        cartesian_modes = np.identity(3 * len(symbols)).reshape(3 * len(symbols), len(symbols), 3)

        total_project = np.zeros_like(cartesian_modes)
        for angle in [2 * np.pi / self._order * self._exp, -2 * np.pi / self._order * self._exp]:

            # angle = 2 * np.pi / self._order * self._exp
            operation = rotation(angle, rotated_axis)

            permu = self._get_permutation(operation, coordinates, symbols)

            projected_modes = []
            for i, mode in enumerate(cartesian_modes):
                operated_mode = np.dot(operation, np.array(mode).T).T
                projected_modes.append(operated_mode[permu])

            # return np.array(projected_modes)
            total_project += np.array(projected_modes)

        return total_project/2

    def get_measure_pos(self, coordinates, symbols, orientation=None, normalized=True):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        measure_coor = []
        # for angle in np.linspace(2*np.pi/self._order, 2*np.pi, self._order)[:-1]:
        for angle in [2 * np.pi / self._order * self._exp, -2 * np.pi / self._order * self._exp]:
            operation = rotation(angle, rotated_axis)

            permu_coor = self._get_operated_coordinates(operation, coordinates, symbols)
            mesure_coor = np.einsum('ij, ij -> ', coordinates, permu_coor)

            measure_coor.append(mesure_coor)

        measure_coor_total = np.average(measure_coor)

        if normalized:
            measure_coor_total /= np.einsum('ij, ij -> ', coordinates, coordinates)

        return measure_coor_total

    def get_operated_coordinates(self, coordinates, symbols, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        operated_coordinates = []
        for angle in [2 * np.pi / self._order * self._exp, -2 * np.pi / self._order * self._exp]:
            operation = rotation(angle, rotated_axis)
            operated_coordinates.append(self._get_operated_coordinates(operation, coordinates, symbols))

        return np.average(operated_coordinates, axis=0)

    def get_overlap_func(self, op_function1, op_function2, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        angle = 2 * np.pi / self._order * self._exp
        operation = rotation(angle, rotated_axis)

        fn_function_r = op_function1.copy()
        fn_function_r.apply_linear_transformation(operation)

        return (op_function2*fn_function_r).integrate

    def apply_rotation(self, orientation):
        self._axis = orientation.apply(self._axis)

    @property
    def axis(self):
        return self._axis

    @property
    def order(self):
        return self._order

    @property
    def exp(self):
        return self._exp

    @property
    def operation_matrix_list(self):
        return [rotation(angle, self._axis) for angle in
                [2 * np.pi / self._order * self._exp, -2 * np.pi / self._order * self._exp]]

    def __mul__(self, other):
        if not other.__class__.__bases__[0] is Operation:
            raise Exception('Product only defined between Operation subclasses')
        else:
            op_list = []
            for op_mat in other.operation_matrix_list:
                new_axis = np.dot(op_mat, self._axis)
                op_list.append(Rotation(self._label, new_axis, self._order, self._exp))

            return op_list
